import torch
from hypothesis import given, settings, strategies as st
from torchcrf import CRF
from torchnyan import BATCH_SIZE, TOKEN_SIZE, assert_close, assert_grad_close, assert_sequence_close, device, sizes
from torchrua import C, D, P

from torchlatent.crf import CrfDecoder, crf_partitions, crf_scores
from torchlatent.semiring import Log


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    num_targets=sizes(TOKEN_SIZE),
    rua_emissions=st.sampled_from([C.new, D.new, P.new]),
    rua_targets=st.sampled_from([C.new, D.new, P.new]),
)
def test_crf_scores(token_sizes, num_targets, rua_emissions, rua_targets):
    inputs = [
        torch.randn((token_size, num_targets), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    targets = [
        torch.randint(0, num_targets, (token_size,), device=device)
        for token_size in token_sizes
    ]

    expected_crf = CRF(num_tags=num_targets, batch_first=False).to(device=device)

    expected_emissions = D.new(inputs)
    expected_tags = D.new(targets)

    expected = expected_crf._compute_score(
        expected_emissions.data.transpose(0, 1),
        expected_tags.data.transpose(0, 1),
        expected_emissions.mask().transpose(0, 1),
    )

    actual = crf_scores(
        emissions=rua_emissions(inputs),
        targets=rua_targets(targets),
        transitions=(expected_crf.transitions, expected_crf.start_transitions, expected_crf.end_transitions),
        semiring=Log,
    )

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    num_targets=sizes(TOKEN_SIZE),
    rua_emissions=st.sampled_from([C.new, D.new, P.new]),
)
def test_crf_partitions(token_sizes, num_targets, rua_emissions):
    inputs = [
        torch.randn((token_size, num_targets), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected_crf = CRF(num_tags=num_targets, batch_first=False).to(device=device)

    expected_emissions = D.new(inputs)

    expected = expected_crf._compute_normalizer(
        expected_emissions.data.transpose(0, 1),
        expected_emissions.mask().t(),
    )

    actual = crf_partitions(
        emissions=rua_emissions(inputs),
        transitions=(expected_crf.transitions, expected_crf.start_transitions, expected_crf.end_transitions),
        semiring=Log,
    )

    assert_close(actual=actual, expected=expected, rtol=1e-4, atol=1e-4)
    assert_grad_close(actual=actual, expected=expected, inputs=inputs, rtol=1e-4, atol=1e-4)


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TOKEN_SIZE),
    num_targets=sizes(TOKEN_SIZE),
    rua_emissions=st.sampled_from([C.new, D.new, P.new]),
)
def test_crf_argmax(token_sizes, num_targets, rua_emissions):
    inputs = [
        torch.randn((token_size, num_targets), device=device, requires_grad=True)
        for token_size in token_sizes
    ]

    expected_crf = CRF(num_tags=num_targets, batch_first=False).to(device=device)

    expected_emissions = D.new(inputs)

    expected = expected_crf.decode(
        expected_emissions.data.transpose(0, 1),
        expected_emissions.mask().t(),
    )
    expected = C.new([torch.tensor(tensor, device=device) for tensor in expected])

    actual_crf = CrfDecoder(num_targets=num_targets)
    actual_crf.transitions = expected_crf.transitions
    actual_crf.head_transitions = expected_crf.start_transitions
    actual_crf.last_transitions = expected_crf.end_transitions

    actual = actual_crf(rua_emissions(inputs)).argmax.cat()

    assert_sequence_close(actual=actual, expected=expected)



from typing import Tuple, Type, Union

import torch
from torch import Tensor, nn
from torch.distributions.utils import lazy_property
from torch.nn import init
from torchrua import C, D, P

from torchlatent.abc import StructuredDecoder, StructuredDistribution
from torchlatent.semiring import Log, Max, Semiring

T = Tuple[Tensor, Tensor, Tensor]


def crf_scores(emissions: Union[C, D, P], targets: Union[C, D, P], transitions: T, semiring: Type[Semiring]) -> Tensor:
    transitions, head_transitions, last_transitions = transitions

    targets = _, token_sizes = targets.cat()
    head_transitions = targets.head().rua(head_transitions)
    last_transitions = targets.last().rua(last_transitions)
    transitions = targets.data.roll(1).rua(transitions, targets)

    emissions, _ = emissions.idx().cat().rua(emissions, targets)
    emissions = semiring.segment_prod(emissions, sizes=token_sizes)

    token_sizes = torch.stack([torch.ones_like(token_sizes), token_sizes - 1], dim=-1)
    transitions = semiring.segment_prod(transitions, sizes=token_sizes.view(-1))[1::2]

    return semiring.mul(
        semiring.mul(head_transitions, last_transitions),
        semiring.mul(emissions, transitions),
    )


def crf_partitions(emissions: Union[C, D, P], transitions: T, semiring: Type[Semiring]) -> Tensor:
    transitions, head_transitions, last_transitions = transitions

    emissions = emissions.pack()
    last_indices = emissions.idx().last()
    emissions, batch_sizes, _, _ = emissions

    _, *batch_sizes = sections = batch_sizes.detach().cpu().tolist()
    emission, *emissions = torch.split(emissions, sections, dim=0)

    charts = [semiring.mul(head_transitions, emission)]
    for emission, batch_size in zip(emissions, batch_sizes):
        charts.append(semiring.mul(
            semiring.bmm(charts[-1][:batch_size], transitions),
            emission,
        ))

    emission = torch.cat(charts, dim=0)[last_indices]
    return semiring.sum(semiring.mul(emission, last_transitions), dim=-1)


class CrfDistribution(StructuredDistribution):
    def __init__(self, emissions: Union[C, D, P], transitions: T) -> None:
        super(CrfDistribution, self).__init__(emissions=emissions)
        self.transitions = transitions

    def log_scores(self, targets: Union[C, D, P]) -> Tensor:
        return crf_scores(
            emissions=self.emissions, targets=targets,
            transitions=self.transitions,
            semiring=Log,
        )

    @lazy_property
    def log_partitions(self) -> Tensor:
        return crf_partitions(
            emissions=self.emissions,
            transitions=self.transitions,
            semiring=Log,
        )

    @lazy_property
    def max(self) -> Tensor:
        return crf_partitions(
            emissions=self.emissions,
            transitions=self.transitions,
            semiring=Max,
        )

    @lazy_property
    def argmax(self) -> Union[C, D, P]:
        argmax = super(CrfDistribution, self).argmax.argmax(dim=-1)
        return self.emissions._replace(data=argmax)


class CrfDecoder(StructuredDecoder):
    def __init__(self, *, num_targets: int) -> None:
        super(CrfDecoder, self).__init__(num_targets=num_targets)

        self.transitions = nn.Parameter(torch.empty((num_targets, num_targets)))
        self.head_transitions = nn.Parameter(torch.empty((num_targets,)))
        self.last_transitions = nn.Parameter(torch.empty((num_targets,)))

        self.reset_parameters()

    def reset_parameters(self) -> None:
        init.zeros_(self.transitions)
        init.zeros_(self.head_transitions)
        init.zeros_(self.last_transitions)

    def forward(self, emissions: Union[C, D, P]) -> CrfDistribution:
        return CrfDistribution(
            emissions=emissions,
            transitions=(
                self.transitions,
                self.head_transitions,
                self.last_transitions,
            ),
        )



import torch
from hypothesis import given, settings, strategies as st
from torchnyan.assertion import assert_close, assert_grad_close
from torchnyan.strategy import TINY_BATCH_SIZE, TINY_TOKEN_SIZE, device, sizes

from torchlatent.functional import logaddexp, logsumexp


@settings(deadline=None)
@given(
    token_sizes=sizes(TINY_BATCH_SIZE, TINY_TOKEN_SIZE)
)
def test_logaddexp(token_sizes):
    x = torch.randn(token_sizes, device=device, requires_grad=True)
    y = torch.randn(token_sizes, device=device, requires_grad=True)

    actual = logaddexp(x, y)
    expected = torch.logaddexp(x, y)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=(x, y))


@settings(deadline=None)
@given(
    data=st.data(),
    token_sizes=sizes(TINY_BATCH_SIZE, TINY_TOKEN_SIZE)
)
def test_logsumexp(data, token_sizes):
    tensor = torch.randn(token_sizes, device=device, requires_grad=True)
    dim = data.draw(st.integers(min_value=-len(token_sizes), max_value=len(token_sizes) - 1))

    actual = logsumexp(tensor, dim=dim)
    expected = torch.logsumexp(tensor, dim=dim)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=(tensor,))



import torch
from torch import Tensor
from torchrua import segment_logsumexp, segment_max, segment_prod, segment_sum

from torchlatent.functional import logaddexp, logsumexp

__all__ = [
    'Semiring', 'ExceptionSemiring',
    'Std', 'Log', 'Max', 'Xen', 'Div',

]


class Semiring(object):
    zero: float
    one: float

    @classmethod
    def eye_like(cls, tensor: Tensor) -> Tensor:
        *_, n = tensor.size()

        eye = torch.full((n, n), fill_value=cls.zero, dtype=tensor.dtype, device=tensor.device)
        index = torch.arange(n, dtype=torch.long, device=tensor.device)
        eye[index, index] = cls.one
        return eye

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def sum(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        raise NotImplementedError

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        raise NotImplementedError

    @classmethod
    def segment_sum(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def segment_prod(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def bmm(cls, x: Tensor, y: Tensor) -> Tensor:
        return cls.sum(cls.mul(x[..., :, :, None], y[..., None, :, :]), dim=-2, keepdim=False)


class Std(Semiring):
    zero = 0.
    one = 1.

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        return x + y

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        return x * y

    @classmethod
    def sum(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.prod(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def segment_sum(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        return segment_sum(tensor, segment_sizes=sizes)

    @classmethod
    def segment_prod(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        return segment_prod(tensor, segment_sizes=sizes)


class Log(Semiring):
    zero = -float('inf')
    one = 0.

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        return logaddexp(x, y)

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        return x + y

    @classmethod
    def sum(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return logsumexp(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def segment_sum(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        return segment_logsumexp(tensor, segment_sizes=sizes)

    @classmethod
    def segment_prod(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        return segment_sum(tensor, segment_sizes=sizes)


class Max(Semiring):
    zero = -float('inf')
    one = 0.

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        return torch.maximum(x, y)

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        return x + y

    @classmethod
    def sum(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.max(tensor, dim=dim, keepdim=keepdim).values

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def segment_sum(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        return segment_max(tensor, segment_sizes=sizes)

    @classmethod
    def segment_prod(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        return segment_sum(tensor, segment_sizes=sizes)


class ExceptionSemiring(Semiring):
    @classmethod
    def sum(cls, tensor: Tensor, log_p: Tensor, log_q: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        raise NotImplementedError

    @classmethod
    def segment_sum(cls, tensor: Tensor, log_p: Tensor, log_q: Tensor, sizes: Tensor) -> Tensor:
        raise NotImplementedError


class Xen(ExceptionSemiring):
    zero = 0.
    one = 0.

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        return x + y

    @classmethod
    def sum(cls, tensor: Tensor, log_p: Tensor, log_q: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum((tensor - log_q) * log_p.exp(), dim=dim, keepdim=keepdim)

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def segment_sum(cls, tensor: Tensor, log_p: Tensor, log_q: Tensor, sizes: Tensor) -> Tensor:
        return segment_sum((tensor - log_q) * log_p.exp(), segment_sizes=sizes)

    @classmethod
    def segment_prod(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        return segment_sum(tensor, segment_sizes=sizes)


class Div(ExceptionSemiring):
    zero = 0.
    one = 0.

    @classmethod
    def add(cls, x: Tensor, y: Tensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def mul(cls, x: Tensor, y: Tensor) -> Tensor:
        return x + y

    @classmethod
    def sum(cls, tensor: Tensor, log_p: Tensor, log_q: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum((tensor - log_q + log_p) * log_p.exp(), dim=dim, keepdim=keepdim)

    @classmethod
    def prod(cls, tensor: Tensor, dim: int, keepdim: bool = False) -> Tensor:
        return torch.sum(tensor, dim=dim, keepdim=keepdim)

    @classmethod
    def segment_sum(cls, tensor: Tensor, log_p: Tensor, log_q: Tensor, sizes: Tensor) -> Tensor:
        return segment_sum((tensor - log_q + log_p) * log_p.exp(), segment_sizes=sizes)

    @classmethod
    def segment_prod(cls, tensor: Tensor, sizes: Tensor) -> Tensor:
        return segment_sum(tensor, segment_sizes=sizes)



from abc import ABCMeta
from typing import Union

import torch
import torch.autograd
from torch import Tensor, nn
from torch.distributions.utils import lazy_property
from torchrua import C, D, P


class StructuredDistribution(object, metaclass=ABCMeta):
    def __init__(self, emissions: Union[C, D, P]) -> None:
        super(StructuredDistribution, self).__init__()
        self.emissions = emissions

    def log_scores(self, targets: Union[C, D, P]) -> Tensor:
        raise NotImplementedError

    def log_probs(self, targets: Union[C, D, P]) -> Tensor:
        return self.log_scores(targets=targets) - self.log_partitions

    @lazy_property
    def log_partitions(self) -> Tensor:
        raise NotImplementedError

    @lazy_property
    def marginals(self) -> Tensor:
        grad, = torch.autograd.grad(
            self.log_partitions, self.emissions.data, torch.ones_like(self.log_partitions),
            create_graph=True, retain_graph=True, only_inputs=True, allow_unused=True,

        )
        return grad

    @lazy_property
    def max(self) -> Tensor:
        raise NotImplementedError

    @lazy_property
    def argmax(self) -> Tensor:
        grad, = torch.autograd.grad(
            self.max, self.emissions.data, torch.ones_like(self.max),
            create_graph=False, retain_graph=False, only_inputs=True, allow_unused=True,
        )
        return grad


class StructuredDecoder(nn.Module):
    def __init__(self, *, num_targets: int) -> None:
        super(StructuredDecoder, self).__init__()
        self.num_targets = num_targets

    def reset_parameters(self) -> None:
        pass

    def extra_repr(self) -> str:
        return f'num_targets={self.num_targets}'

    def forward(self, emissions: Union[C, D, P]) -> StructuredDistribution:
        raise NotImplementedError



import torch
from hypothesis import given, settings, strategies as st
from torch_struct import TreeCRF
from torchnyan import BATCH_SIZE, TINY_TOKEN_SIZE, assert_close, assert_grad_close, device, sizes
from torchrua import C

from torchlatent.cky import CkyDecoder, cky_partitions, cky_scores
from torchlatent.semiring import Log


def get_argmax(cky):
    argmax = cky.argmax
    mask = argmax > 0

    _, t, _, n = mask.size()
    index = torch.arange(t, device=mask.device)
    x = torch.masked_select(index[None, :, None, None], mask=mask)
    y = torch.masked_select(index[None, None, :, None], mask=mask)

    index = torch.arange(n, device=mask.device)
    z = torch.masked_select(index[None, None, None, :], mask=mask)

    return argmax, x, y, z


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    num_targets=sizes(TINY_TOKEN_SIZE),
    rua_targets=st.sampled_from([C.cat, C.pad, C.pack]),
)
def test_cky_scores(token_sizes, num_targets, rua_targets):
    emissions = torch.randn(
        (len(token_sizes), max(token_sizes), max(token_sizes), num_targets),
        device=device, requires_grad=True,
    )
    token_sizes = torch.tensor(token_sizes, device=device)
    expected_cky = TreeCRF(emissions, lengths=token_sizes)

    argmax, x, y, z = get_argmax(expected_cky)

    emissions = torch.randn_like(emissions, requires_grad=True)

    expected_cky = TreeCRF(emissions, lengths=token_sizes)
    expected = expected_cky.log_prob(argmax) + expected_cky.partition

    targets = C(data=torch.stack([x, y, z], dim=-1), token_sizes=token_sizes * 2 - 1)
    actual = cky_scores(
        emissions=C(emissions, token_sizes),
        targets=rua_targets(targets),
        semiring=Log,
    )

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=(emissions,))


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    num_targets=sizes(TINY_TOKEN_SIZE),
)
def test_cky_partitions(token_sizes, num_targets):
    emissions = torch.randn(
        (len(token_sizes), max(token_sizes), max(token_sizes), num_targets),
        device=device, requires_grad=True,
    )
    token_sizes = torch.tensor(token_sizes, device=device)

    expected = TreeCRF(emissions, lengths=token_sizes).partition

    actual_emissions = C(
        data=emissions.logsumexp(dim=-1),
        token_sizes=token_sizes,
    )
    actual = cky_partitions(actual_emissions, Log)

    assert_close(actual=actual, expected=expected)
    assert_grad_close(actual=actual, expected=expected, inputs=(emissions,))


@settings(deadline=None)
@given(
    token_sizes=sizes(BATCH_SIZE, TINY_TOKEN_SIZE),
    num_targets=sizes(TINY_TOKEN_SIZE),
)
def test_cky_argmax(token_sizes, num_targets):
    emissions = torch.randn(
        (len(token_sizes), max(token_sizes), max(token_sizes), num_targets),
        device=device, requires_grad=True,
    )
    token_sizes = torch.tensor(token_sizes, device=device)

    expected_cky = TreeCRF(emissions, lengths=token_sizes)

    _, x, y, z = get_argmax(expected_cky)

    expected = C(data=torch.stack([x, y, z], dim=-1), token_sizes=token_sizes * 2 - 1)

    actual_cky = CkyDecoder(num_targets=num_targets)
    actual = actual_cky(emissions=C(emissions, token_sizes)).argmax

    for actual, expected in zip(actual.tolist(), expected.tolist()):
        assert set(map(tuple, actual)) == set(map(tuple, expected))



from typing import Tuple, Type, Union

import torch
from torch import Tensor
from torch.distributions.utils import lazy_property
from torchrua import C, D, P

from torchlatent.abc import StructuredDecoder, StructuredDistribution
from torchlatent.semiring import Log, Max, Semiring


def cky_scores(emissions: C, targets: Union[C, D, P], semiring: Type[Semiring]) -> Tensor:
    xyz, token_sizes = targets = targets.cat()
    batch_ptr, _ = targets.ptr()

    emissions = emissions.data[batch_ptr, xyz[..., 0], xyz[..., 1], xyz[..., 2]]
    return semiring.segment_prod(emissions, token_sizes)


def diag(tensor: Tensor, offset: int) -> Tensor:
    return tensor.diagonal(offset=offset, dim1=1, dim2=2)


def diag_scatter(chart: Tensor, score: Tensor, offset: int) -> None:
    chart.diagonal(offset=offset, dim1=1, dim2=2)[::] = score


def left(chart: Tensor, offset: int) -> Tensor:
    b, t, _, *size = chart.size()
    c, n, m, *stride = chart.stride()
    return chart.as_strided(
        size=(b, t - offset, offset, *size),
        stride=(c, n + m, m, *stride),
    )


def right(chart: Tensor, offset: int) -> Tensor:
    b, t, _, *size = chart.size()
    c, n, m, *stride = chart.stride()
    return chart[:, 1:, offset:].as_strided(
        size=(b, t - offset, offset, *size),
        stride=(c, n + m, n, *stride),
    )


def cky_partitions(emissions: C, semiring: Type[Semiring]) -> Tensor:
    chart = torch.full_like(emissions.data, fill_value=semiring.zero, requires_grad=False)

    diag_scatter(chart, diag(emissions.data, offset=0), offset=0)

    for w in range(1, chart.size()[1]):
        score = semiring.sum(semiring.mul(left(chart, offset=w), right(chart, offset=w)), dim=2)
        diag_scatter(chart, semiring.mul(score, diag(emissions.data, offset=w)), offset=w)

    index = torch.arange(chart.size()[0], dtype=torch.long, device=chart.device)
    return chart[index, 0, emissions.token_sizes - 1]


def masked_select(mask: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
    _, t, _, n = mask.size()

    index = torch.arange(t, device=mask.device)
    x = torch.masked_select(index[None, :, None, None], mask=mask)
    y = torch.masked_select(index[None, None, :, None], mask=mask)

    index = torch.arange(n, device=mask.device)
    z = torch.masked_select(index[None, None, None, :], mask=mask)

    return x, y, z


class CkyDistribution(StructuredDistribution):
    def __init__(self, emissions: C) -> None:
        super(CkyDistribution, self).__init__(emissions=emissions)

    def log_scores(self, targets: Union[C, D, P]) -> Tensor:
        return cky_scores(
            emissions=self.emissions, targets=targets,
            semiring=Log,
        )

    @lazy_property
    def log_partitions(self) -> Tensor:
        return cky_partitions(
            emissions=self.emissions._replace(data=Log.sum(self.emissions.data, dim=-1)),
            semiring=Log,
        )

    @lazy_property
    def max(self) -> Tensor:
        return cky_partitions(
            emissions=self.emissions._replace(data=Max.sum(self.emissions.data, dim=-1)),
            semiring=Max,
        )

    @lazy_property
    def argmax(self) -> C:
        argmax = super(CkyDistribution, self).argmax
        x, y, z = masked_select(argmax > 0)

        return C(
            data=torch.stack([x, y, z], dim=-1),
            token_sizes=self.emissions.token_sizes * 2 - 1,
        )


class CkyDecoder(StructuredDecoder):
    def __init__(self, *, num_targets: int) -> None:
        super(CkyDecoder, self).__init__(num_targets=num_targets)

    def forward(self, emissions: C) -> CkyDistribution:
        return CkyDistribution(emissions=emissions)
